
from Vectors import Vector
# from Path import Path
from Arc import *

import math

# missing: hook arrows

sw = 1
hw = 3.6
# A, B always in radians
A = 24*math.pi/180.0
B = 60*math.pi/180.0

""" 

An arrow is made up of three components:
tail, shaft, head, which are joined together
at certain `gates'.  Anything with two
gates sw apart can be a head or tail,
andything with top and bottom
paths leading to and from two gates can be a shaft.
Heads and tails are attached to shafts at the gates.
Arrows can also be left open.

Tail: start, stop
Head: start, stop
Shaft: p3, p0 is left gate, p1, p2 the other
(This is bad notation, and should probably be changed.)

Every head also has a nominal length, x.length.
Because all transformations of x are rigid,
this doesn't change.  A head's length is to be used
to design the shaft so as to hit 
the endpoint reasonably closely.

Each arrow exists in a basic normalized place
which can be shifted with parameters p (origin) and a (angle).
Each should give rise to a path assembled from components.
Then these can be transformed easily by translation
and rotating.

"""

class PlainHead:
	# This is the basic arrowhead
	# origin at (0,0)
	def __init__(self):
		global sw
		global hw
		global A
		global B
		# builds the path
		# xA = horizontal distance from tip to wing tip
		xA = 0.5*hw/math.tan(A)
		# xB = horizontal distance from p to wing tip
		xB = 0.5*(hw-sw)/math.tan(B)
		# path starts at [0, -half shaft width]
		self.p0 = Vector([0, -0.5*sw])
		self.p1 = self.p0 + [-xB, -0.5*(hw-sw)]
		self.p2 = self.p1 + [ xA,  0.5*hw]
		self.p3 = self.p2 + [-xA,  0.5*hw]
		self.p4 = self.p3 + [ xB, -0.5*(hw-sw)]
		self.length = xA - xB

	# rotates by a and then translates by p
	def translate(self, p):
		Arrow.Translate(self.p0, p)
		Arrow.Translate(self.p1, p)
		Arrow.Translate(self.p2, p)
		Arrow.Translate(self.p3, p)
		Arrow.Translate(self.p4, p)
		
	def rotate(self, a):
		c = math.cos(a); s = math.sin(a)
		Arrow.Rotate(self.p0, c, s)
		Arrow.Rotate(self.p1, c, s)
		Arrow.Rotate(self.p2, c, s)
		Arrow.Rotate(self.p3, c, s)
		Arrow.Rotate(self.p4, c, s)
		
	def start(self):
		return self.p0
	
	def stop(self):
		return self.p4

	def beginpath(self, p):
		p.moveto(self.p0)

	# assumes p0 = current point
	def mkpath(self, p):
		p.lineto(self.p1)
		p.lineto(self.p2)
		p.lineto(self.p3)
		p.lineto(self.p4)

class StraightShaft:
	# origin = (0,0)
	# L = length
	def __init__(self, L):
		global sw
		self.p0 = [ 0, -0.5*sw ]
		self.p1 = [ L, -0.5*sw ]
		self.p2 = [ L,  0.5*sw ]
		self.p3 = [ 0,  0.5*sw ]

	# rotates by a and then translates by p
	def translate(self, p):
		Arrow.Translate(self.p0, p)
		Arrow.Translate(self.p1, p)
		Arrow.Translate(self.p2, p)
		Arrow.Translate(self.p3, p)
		
	def rotate(self, a):
		c = math.cos(a); s = math.sin(a)
		Arrow.Rotate(self.p0, c, s)
		Arrow.Rotate(self.p1, c, s)
		Arrow.Rotate(self.p2, c, s)
		Arrow.Rotate(self.p3, c, s)
		
	# assumes p2 = current point
	def mkbottompath(self, p):
		p.lineto(self.p1)

	# assumes p0 = current point
	def mktoppath(self, p):
		p.lineto(self.p3)
		
class ArcShaft:
	# origin = (0,0)
	# basic one: arc of radius r from angle 0 to A
	def __init__(self, A, r):
		global sw
		c = math.cos(A)
		s = math.sin(A)
		e = 0.5*sw
		self.p0 = Vector([r + e, 0])
		self.p3 = Vector([r - e, 0])
		self.p1 = Vector([(r+e)*c, (r+e)*s])
		self.p2 = Vector([(r-e)*c, (r-e)*s])
		self.A0 = 0
		self.A1 = A
		self.r = r
		self.ctr = Vector([0,0])
	
	# rotates by a and then translates by p
	def translate(self, p):
		Arrow.Translate(self.p0, p)
		Arrow.Translate(self.p1, p)
		Arrow.Translate(self.p2, p)
		Arrow.Translate(self.p3, p)
		Arrow.Translate(self.ctr, p)
		
	def rotate(self, a):
		c = math.cos(a); s = math.sin(a)
		Arrow.Rotate(self.p0, c, s)
		Arrow.Rotate(self.p1, c, s)
		Arrow.Rotate(self.p2, c, s)
		Arrow.Rotate(self.p3, c, s)
		Arrow.Rotate(self.ctr, c, s)
		self.A0 += a
		self.A1 += a
	
	def mkbottompath(self, p):
		p.arc(self.ctr, self.r + 0.5*sw, self.A0, self.A1)

	# assumes p0 = current point
	def mktoppath(self, p):
		p.arcn(self.ctr, self.r - 0.5*sw, self.A1, self.A0)
		
	def start(self):
		return self.p0
		
class ArcnShaft(ArcShaft):

	def __init__(self, A, r):
		ArcShaft.__init__(self, A, r)
		t = self.p0
		self.p0 = self.p2
		self.p2 = t
		t = self.p1
		self.p1 = self.p3
		self.p3 = t

	# assumes p2 = current point
	def mkbottompath(self, p):
		p.arcn(self.ctr, self.r - 0.5*sw, self.A1, self.A0)

	# assumes p0 = current point
	def mktoppath(self, p):
		p.arc(self.ctr, self.r + 0.5*sw, self.A0, self.A1)
		
	def start(self):
		return self.p0
		
class StubTail:
	# sw, cap mode, angle
	def __init__(self):
		global sw
		self.p0 = Vector([0,  0.5*sw])
		self.p1 = Vector([0, -0.5*sw])

	def translate(self, p):
		Arrow.Translate(self.p0, p)
		Arrow.Translate(self.p1, p)

	def rotate(self, a):
		c = math.cos(a); s = math.sin(a)
		Arrow.Rotate(self.p0, c, s)
		Arrow.Rotate(self.p1, c, s)
		
	def start(self):
		return self.p0

	def stop(self):
		return self.p1
	
	# assumes p0 = current point
	# p = Path, Canvas: commands moveto etc.
	# NO: a Path is not active, has no coordinate system
	def mkpath(self, p):
		global sw
		p0 = self.p0
		p1 = self.p1
		c = p.currentlinecap()
		if c == 0:
			p.lineto(p1)
		elif c == 1:
			ctr = (p0 + p1)*0.5
			A = math.atan2(p0[1]-p1[1], p0[0]-p1[0])
			p.arc(ctr[0], ctr[1], 0.5*sw, A, A + math.pi)
			""" a = Arc(ctr[0], ctr[1], 0.5*sw, A, A + math.pi)
			a.mkpath(p) """
		else: # c == 2:
			p0 = self.p0
			p1 = self.p1
			v = (p0 - p1)*0.5
			p.lineto(p0 + [-v[1], v[0]])
			p.lineto(p1 + [-v[1], v[0]])
			p.lineto(p1)

class Arrow:
	# tail might be None
	def __init__(self, head, shaft, tail):
		# shaft = StraightShaft(L-head.length)
		Arrow.attach(head, head.start(), head.stop(), shaft.p1, shaft.p2)
		if not tail == None:
			Arrow.attach(tail, tail.start(), tail.stop(), shaft.p3, shaft.p0)
		self.shaft = shaft
		self.head = head
		self.tail = tail
	
	def start(self):
		return self.shaft.p0
		
	# rotates by a and then translates by p
	def translate(self, p):
		self.shaft.translate(p)
		self.head.translate(p)
		if not self.tail == None:
			self.tail.translate(p)
		
	def rotate(self, a):
		self.shaft.rotate(a)
		if not self.head == None:
			self.head.rotate(a)
		if not self.tail == None:
			self.tail.rotate(a)
		
	def mkpath(self, p):
		p.moveto(self.shaft.p0)
		self.shaft.mkbottompath(p)
		self.head.mkpath(p)
		self.shaft.mktoppath(p)
		if not self.tail == None:
			self.tail.mkpath(p)
			p.closepath()
		
	# === static routines ============================================
		
	def intersection(p, v, P, V):
		# intersection: P-t*V lies on line p + tv or 
		# A.x+B.y = A.p[0]+B.p[1] = C is the equation of v + tv
		A = -v[1]
		B =  v[0]
		C = A*p[0]+B*p[1]
		# A*(P[0]-t*V[0])+B*(P[1]-t*V[1]) = C 
		t = (A*P[0]+B*P[1]-C)/float(A*V[0]+B*V[1])
		Q = [ P[0]-t*V[0], P[1]-t*V[1] ]
		return(Q)
	intersection = staticmethod(intersection)

	def vectorlength(dx, dy):
		dx = float(abs(dx)); dy = float(abs(dy))
		if dx > dy:
			ds = dy/dx
			return dx*math.sqrt(1 + ds*ds)
		ds = dx/dy
		return dy*math.sqrt(1 + ds*ds)
	vectorlength = staticmethod(vectorlength)

	def Rotate(p, c, s):
		x = p[0]
		y = p[1]
		p[0] = c*x - s*y
		p[1] = s*x + c*y
	Rotate = staticmethod(Rotate)

	def Translate(p, v):
		p[0] += v[0]
		p[1] += v[1]
	Translate = staticmethod(Translate)
	
	def subdivide(P):
		p0 = P[0]
		p1 = P[1]
		p2 = P[2]
		p01 = [ 0.5*(p0[0]+p1[0]), 0.5*(p0[1]+p1[1]) ]
		p12 = [ 0.5*(p2[0]+p1[0]), 0.5*(p2[1]+p1[1]) ]
		p012 = [ 0.5*(p01[0]+p12[0]), 0.5*(p01[1]+p12[1]) ]
		return [
			[ p0, p01, p012 ],
			[ p012, p12, p2 ]
		]
	subdivide = staticmethod(subdivide)
	
	# these position the start and stop to match the shaft
	def attach(h, t0, t1, s0, s1):
		h.translate([-t0[0], -t0[1]])
		u = [ s1[0]-s0[0], s1[1]-s0[1] ]
		v = [ t1[0]-t0[0], t1[1]-t0[1] ]
		a = math.atan2(u[1], u[0])
		b = math.atan2(v[1], v[0])
		h.rotate(a-b)
		h.translate([s0[0], s0[1]])
	attach = staticmethod(attach)

"""
The head has length dl, starting at some point (r cos(T), r sin(T) )
and goes tangentially to the line where sin(T) + dl cos(T) e angle is B.  we want to know
T, which is the end of the arc.  So
the angle of ( r cos(T) - dl sin(T), r sin(T) + dl cos(T) ) is B.
So

	sin(B) / cos (B) = r sin(T) + dl cos(T) /  r cos(T) - dl sin(T)
	
We solve this for tan of T, knowing that T is close to B.
	
Similarly, the quad arrow should follow the given 
quad arcs, stop in time to fit head on with tangential direction.

At intermediate points P0, P1, P2 we calculate P01 = P1-P0, P12 = P2-P1,
then unit vectors u perp to P01, pointing left; v perp to P12, also pointing left.
We then want to find a, b such that the projection of au+bv has projection wd onto both u, v:
(au+bv)*u = sw/2, (au+bv)*v = sw/2.

	a, b = (wd/(u*v)^2-1) . (u*v-1, 1-u*v)


The curve that stops at P(t) is the Bezier quadratic
with Q0 = P0, Q1 = (1-t)P0 + t P1, Q2 = P(t); if s = 1-t
P(t) = P2 - 2 s (P2-P1) + s^{2}((P2-P1) - (P1-P0)). 
Let P12 = P2-P1, P01 = P1-P0.
Let b P12 of P01 onto P12.  Then
For s near 0 this is approximately
P2 - 2 s P12 + s^{2}(1 - b) P12 = P2 - (P2-P1)(2s - a s^2)
where a = 1 - b, which is on the line [P1,P2].  
The point at distance
d from P2 on this line corresponds to 2s - a s^2 = d/|P2-P1| = d/D,
s = (1/a)(1 - sqrt(1 - da/D)) = (d/D) / 1 + sqrt(1 - da/D).
 = 1 - d/D, t = 1-s = sqrt(1 - d/D).
 if a = 0, this means t = sqrt(1 - d/D),
s = 1 - t = 1 - sqrt(1 - d/D) = (d/D) / (1 + t). So
the endpoint of our quadratic curve is P(t) for this value of t.
A straight line segment of length d from there on,
with velocity P'(t) = 2((1-t)(P1-P0) + t(P2-P1))

""" 

class QuadShaft:

	# args = 2k+1 control points 
	# ell = length of head
	# output = array of bottom control points, of top ones
	# the segments are compatible in that
	# the tengent directions at end of one are the same
	# as those of start of the next
	def __init__(self, ell, P):
		# print "quad shaft"
		global sw
		wd = 0.5*sw
		k = (len(P)-1)/2
		# assemble for each segment three points on each side
		bottom = []
		top = []
		P0 = P[0]
		P1 = P[1]
		P01 = [ P1[0] - P0[0], P1[1] - P0[1] ]
		d01 = Arrow.vectorlength(P01[0], P01[1])
		u = [ -P01[1]/d01, P01[0]/d01 ]
		self.p0 = [ P0[0] - wd*u[0], P0[1] - wd*u[1] ]
		self.p3 = [ P0[0] + wd*u[0], P0[1] + wd*u[1] ]
		bottom.append(self.p0)
		top.append(self.p3)
		for i in range(k-1):
			P2 = P[2*i+2] # last = 2k-2
			P12 = [ P2[0] - P1[0], P2[1] - P1[1] ]
			d12 = Arrow.vectorlength(P12[0], P12[1])
			v = [ -P12[1]/d12, P12[0]/d12 ]
			# u = unit vector left from P0 -> P1
			# v = unit vector left from P1 -> P2
			# w.u = sw/2, e.v = sw/2
			dot = u[0]*v[0]+u[1]*v[1]
			det = dot*dot-1
			a = (wd/det)*(dot-1); b = a
			w = [ a*u[0]+b*v[0], a*u[1]+b*v[1] ]
			# print "w =", w
			# print [ P1[0] - w[0], P1[1] - w[1] ], [ P1[0] + w[0], P1[1] + w[1] ]
			bottom.append([ P1[0] - w[0], P1[1] - w[1] ])
			top.append([ P1[0] + w[0], P1[1] + w[1] ])
			P0 = P2
			P1 = P[2*i+3]
			u = v
			bottom.append([ P0[0] - wd*u[0], P0[1] - wd*u[1] ])
			top.append([ P0[0] + wd*u[0], P0[1] + wd*u[1] ])
		# now the last segment: two more points to add
		P2 = P[-1]
		dx = P2[0]-P1[0]; dy = P2[1]-P1[1]
		D = Arrow.vectorlength(dx, dy)
		P01 = [ P1[0]-P0[0], P1[1]-P0[1] ]
		P12 = [ dx, dy ]
		b = (P01[0]*dx+P01[1]*dy)/float(dx*dx+dy*dy)
		# b = coefficient pf projection of P01 onto P12
		a = 1 - b
		s = (ell/D)/ (1 + math.sqrt(1 - a*ell/D))
		t = 1-s
		P012 = [ P12[0]-P01[0], P12[1]-P01[1] ]
		Q1 = [ s*P0[0]+t*P1[0], s*P0[1]+t*P1[1] ] # middle control point
		Q2 = [ 
			P2[0] - 2*s*P12[0] + s*s*P012[0],
			P2[1] - 2*s*P12[1] + s*s*P012[1] 
		] # endpoint
		# print "q2 =", Q2
		V = [ 2*(P01[0]+t*P012[0]), 2*(P01[1]+t*P012[1]) ]
		D = Arrow.vectorlength(V[0], V[1])
		v = [ -V[1]/D, V[0]/D ] # unit perp at end
		dot = u[0]*v[0]+u[1]*v[1]
		det = dot*dot-1
		a = (wd/det)*(dot-1); b = a
		w = [ a*u[0]+b*v[0], a*u[1]+b*v[1] ]
		# print "w =", w
		p1 = [ Q1[0] - w[0], Q1[1] - w[1] ]
		q1 = [ Q1[0] + w[0], Q1[1] + w[1] ]
		bottom.append(p1)
		top.append(q1)
		self.p1 = [ Q2[0] - wd*v[0], Q2[1] - wd*v[1] ]
		self.p2 = [ Q2[0] + wd*v[0], Q2[1] + wd*v[1] ]
		bottom.append(self.p1)
		top.append(self.p2)
		self.bottom = bottom
		self.top = []
		n = len(top)
		for i in range(n):
			self.top.append(top[n-1-i])
		t = self.top
		b = self.bottom
		n = len(t)
		""" for i in range(n):
			print b[i]
		for i in range(n):
			print t[n-1-i]
		print "-------------------------" """
	
	def mkbottompath(self, ps):
		b = self.bottom
		i = 1
		while i < len(b):
			ps.quadto(b[i], b[i+1])
			i += 2
		
	def mktoppath(self, ps):
		t = self.top
		i = 1
		while i < len(t):
			ps.quadto(t[i], t[i+1])
			i += 2
			
	def rotate(self, a):
		b = self.bottom
		t = self.top
		n = len(b)
		for i in range(n):
			Arrow.rotate(b[i], a)
			Arrow.rotate(t[i], a)
		
	def translate(self, p):
		b = self.bottom
		t = self.top
		n = len(b)
		for i in range(n):
			Arrow.translate(b[i], p)
			Arrow.translate(t[i], p)
		
# TexHead: arc wings; width of wings = r, radius of tip = rho;
# these and many variables (A, B) should be globally fixed 
# so as to match Knuth's arrows; others (R) should scale to hw

class TexHead:
	# sw, hw, origin, angle
	def __init__(self, A, B, R):
		global sw
		global hw
		self.A = A
		self.B = B
		self.R = R
		self.P = [ 0, -0.5*sw ]
		self.Q = [0, 0.5*sw]
		self.axis = [ 1, 0 ]
		r = 0.84*sw
		self.r = r 
		self.rho = 0.8*r
		# more parameters: 
		# 0.8 = ratio of sw's; rho = radius of tip
		c = math.cos(A)
		Ax = math.acos((R*c-sw/2)/(R+r/2))
		self.length = (R+r/2)*math.sin(Ax)-R*math.sin(A)

	def mkpath(self, ps):
		A = self.A
		axis = self.axis
		O = [ 0.5*(self.P[0]+self.Q[0]), 0.5*(self.P[1]+self.Q[1]) ]
		R = self.R
		s = math.sin(A)
		c = math.cos(A)
		a = math.atan2(axis[1], axis[0])
		e = [ R*s*axis[0], R*s*axis[1] ]
		f = [ -R*c*axis[1], R*c*axis[0] ]
		C = [ O[0] + e[0] - f[0], O[1] + e[1] - f[1] ]
		Ax = math.acos((R*c-sw/2)/(R+self.r/2))
		T = math.pi/2+A+a+self.B
		ps.arc(C, R+self.r/2, math.pi/2+Ax+a, T)
		ps.arc(C[0]+R*math.cos(T), C[1]+R*math.sin(T), self.r/2, T, T+math.pi)
		ps.arcn(C, R-self.r/2, T, math.pi/2+A+a) 
		dx = 0.5*(self.r-self.rho)/s
		ps.arc(O[0]+dx*axis[0], O[1]+dx*axis[1], self.rho/2, -math.pi/2+a+A, math.pi/2+a-A)
		C = [ O[0] + e[0] + f[0], O[1] + e[1] + f[1] ]
		T = -math.pi/2-A-self.B+a
		ps.arcn(C, R - self.r/2, -math.pi/2-A+a, T)
		ps.arc(C[0]+R*math.cos(T), C[1]+R*math.sin(T), self.r/2, T-math.pi, T)
		ps.arc(C, R + self.r/2, T, -math.pi/2-Ax+a)
		
	def start(self):
		X = self.length; a = self.axis
		return [ self.P[0] - X*a[0], self.P[1] - X*a[1] ]
	
	def stop(self):
		X = self.length; a = self.axis
		return [ self.Q[0] - X*a[0], self.Q[1] - X*a[1] ]

	def translate(self, p):
		Arrow.Translate(self.P, p)
		Arrow.Translate(self.Q, p)
	
	def rotate(self, a):
		c = math.cos(a); s = math.sin(a)
		Arrow.Rotate(self.P, c, s)
		Arrow.Rotate(self.Q, c, s)
		Arrow.Rotate(self.axis, c, s)
		
		
# -----------------------------------------------------
# -----------------------------------------------------
# -----------------------------------------------------


"""
	sw 
	sw hw
	sw hw A B
"""

def setarrowdims(ps, *args):
	global sw
	global hw
	global A
	global B
	if len(args) == 1:
		sw = float(args[0])
		hw = 3.6*sw
	elif len(args) == 2:
		sw = float(args[0])
		hw = float(args[1])
	else: # 4 arguments
		sw = float(args[0])
		hw = float(args[1])
		A = args[2]*ps.toRad
		B = args[3]*ps.toRad
		
# ---------------------------------------------------

"""	
	[x, y]
	x y
"""

def arrow(ps, *args):
	n = len(args)
	if n == 1: # a single array
		x0 = 0; y0 = 0
		x1 = args[0][0]
		y1 = args[0][1]
	else: # a pair
		a0 = args[0]
		a1 = args[1]
		if Vector.isarray(a0) and Vector.isarray(a1):
			x0 = a0[0]
			y0 = a0[1]
			x1 = a1[0]
			y1 = a1[1]
		else: # a pair of numbers
			x0 = 0; y0 = 0
			x1 = a0
			y1 = a1
	h = PlainHead()
	L = Arrow.vectorlength(x1-x0, y1-y0)
	s = StraightShaft(L-h.length)
	t = StubTail()
	a = Arrow(h, s, t)
	C = math.atan2(y1-y0, x1-x0)
	a.rotate(C)
	a.translate([x0, y0])
	a.mkpath(ps)

def arrowhead(ps, A):
	h = PlainHead()
	h.rotate(A)
	h.beginpath(ps)
	h.mkpath(ps)

def openarrow(ps, *args):
	n = len(args)
	if n == 1:
		x = args[0][0]
		y = args[0][1]
	else:
		x = args[0]
		y = args[1]
	h = PlainHead()
	L = Arrow.vectorlength(x, y)
	s = StraightShaft(L-h.length)
	t = None
	a = Arrow(h, s, t)
	C = math.atan2(y, x)
	a.rotate(C)
	a.mkpath(ps)

def arcarrow(ps, *args):
	if len(args) == 4:
		C = args[0]
		r = args[1]
		A = args[2]
		B = args[3]
	elif len(args) == 5:
		C = (args[0], args[1])
		r = args[2]
		A = args[3]
		B = args[4]
	else:
		print "Wrong number of arguments to arcarrow!"
		sys.exit(1)
	while B < A:
		B += 2*math.pi
	head = PlainHead()
	S = head.length/r
	tail = StubTail()
	dT = B-A
	s = math.sin(dT)
	c = math.cos(dT)
	T = math.atan2(s - S*c, c + S*s)
	shaft = ArcShaft(T, r)
	a = Arrow(head, shaft, tail)
	a.rotate(A)
	a.translate(C)
	a.mkpath(ps)

import sys

def arcnarrow(ps, *args):
	if len(args) == 4:
		C = args[0]
		r = args[1]
		A = args[2]
		B = args[3]
	elif len(args) == 5:
		C = (args[0], args[1])
		r = args[2]
		A = args[3]
		B = args[4]
	else:
		print "Wrong number of arguments to arcnarrow!"
		sys.exit(1)
	while A < B:
		A += 2*math.pi
	head = PlainHead()
	S = head.length/r # tangent of the angle spanned by the head
	tail = StubTail()
	dT = A-B
	s = math.sin(dT)
	c = math.cos(dT)
	T = math.atan2(s - S*c, c + S*s)
	shaft = ArcnShaft(T, r)
	a = Arrow(head, shaft, tail)
	a.rotate(B+math.atan(S))
	a.translate(C)
	a.mkpath(ps)

"""
# in all cases, the arrow follows the path the arc or arcn would
def arcarrow(ps, *args):
	if len(args) > 3:
		# P, r, a, b
		if len(args) > 4:
			P = (args[0], args[1]); r = args[2]; a = args[3]; b = args[4]
		else:
			P = args[0]; r = args[1]; a = args[2]; b = args[3]
		return Arcarrow(ps, P, r, a, b)
	P = args[0]; Q = args[1]; r = args[2]
	global sw, hw, A, B	
	r = float(r)
	dx = float(Q[0]-P[0])
	dy = float(Q[1]-P[1])
	d = Arrow.vectorlength(dx, dy)
	w = [ -dy/d, dx/d ]
	dr = d/r
	c = math.sqrt(1 - dr*dr/4)
	if r < 0:
		c = -c
		r = -r
	Ax = 2*math.acos(c)
	c *= r
	# Ax = total arc from P to Q
	C = [ 0.5*(P[0]+Q[0]) + c*w[0], 0.5*(P[1]+Q[1])+c*w[1] ]
	# C is the center of the arc
	# now find the true arc end
	head = PlainHead()
	cos = math.cos(Ax)
	sin = math.sin(Ax)
	dl = head.length
	T = math.atan2( sin - (dl/r)*cos, cos + (dl/r)*sin )
	shaft = ArcShaft(T, r)
	tail = StubTail()
	a = Arrow(head, shaft, tail)
	rp = [ P[0]-C[0], P[1]-C[1] ]
	S = math.atan2(rp[1], rp[0])
	a.rotate(S)
	a.translate(C) 
	a.mkpath(ps)
	
def arcnarrow(ps, *args):
	if len(args) > 3:
		# P, r, a, b
		if len(args) > 4:
			P = (args[0], args[1]); r = args[2]; a = args[3]; b = args[4]
		else:
			P = args[0]; r = args[1]; a = args[2]; b = args[3]
		return Arcnarrow(ps, P, r, a, b)
	P = args[0]; Q = args[1]; r = args[2]
	global sw, hw, A, B	
	r = float(r)
	dx = float(P[0]-Q[0])
	dy = float(P[1]-Q[1])
	d = Arrow.vectorlength(dx, dy)
	w = [ -dy/d, dx/d ]
	dr = d/r
	c = math.sqrt(1 - dr*dr/4)
	if r < 0:
		c = -c
		r = -r
	Ax = 2*math.acos(c)
	c *= r
	# Ax = total arc from P to Q
	C = [ 0.5*(P[0]+Q[0]) + c*w[0], 0.5*(P[1]+Q[1])+c*w[1] ]
	# C is the center of the arc
	# now find the true arc end
	head = PlainHead()
	cos = math.cos(Ax)
	sin = math.sin(Ax)
	dl = head.length
	T = math.atan2( sin - (dl/r)*cos, cos + (dl/r)*sin )
	shaft = ArcnShaft(T, r)
	shaft.rotate(Ax-T)
	tail = StubTail()
	a = Arrow(head, shaft, tail)
	rp = [ Q[0]-C[0], Q[1]-C[1] ]
	S = math.atan2(rp[1], rp[0])
	a.rotate(S)
	a.translate(C) 
	a.mkpath(ps)

def Arcarrow(ps, c, r, a, b):
	ca = math.cos(a); sa = math.sin(a)
	cb = math.cos(b); sb = math.sin(b)
	P = [ c[0] + r*ca, c[1] + r*sa ]
	Q = [ c[0] + r*cb, c[1] + r*sb ]
	print [P, Q, r]
	arcarrow(ps, P, Q, r)

def Arcnarrow(ps, c, r, a, b):
	ca = math.cos(a); sa = math.sin(a)
	cb = math.cos(b); sb = math.sin(b)
	P = [ c[0] + r*ca, c[1] + r*sa ]
	Q = [ c[0] + r*cb, c[1] + r*sb ]
	arcnarrow(ps, P, Q, r)
"""

def texarrow(ps, *args):
	global hw
	n = len(args)
	if n == 1: # a single array
		x0 = 0; y0 = 0
		x1 = args[0][0]
		y1 = args[0][1]
	else: # a pair
		a0 = args[0]
		a1 = args[1]
		if Vector.isarray(a0) and Vector.isarray(a1):
			x0 = a0[0]
			y0 = a0[1]
			x1 = a1[0]
			y1 = a1[1]
		else: # a pair of numbers
			x0 = 0; y0 = 0
			x1 = a0
			y1 = a1
	ell = Arrow.vectorlength(x1-x0, y1-y0)
	head = TexHead(math.pi/16, math.pi/2.4, 1*hw)
	shaft = StraightShaft(ell-head.length)
	tail = StubTail()
	a = Arrow(head, shaft, tail)
	a.rotate(math.atan2(y1-y0, x1-x0))
	a.translate([x0, y0])
	a.mkpath(ps)

def quadarrow(ps, *args):
	# the shaft will be itself an assembly of shafts
	# args is a quad path
	global sw
	global hw
	if len(args) == 1:
		A = args[0]
	else:
		A = args
	# A = list of control points of the arrow's spine
	head = PlainHead()
	ell = head.length
	shaft = QuadShaft(ell, A)
	tail = StubTail()
	a = Arrow(head, shaft, tail)
	a.mkpath(ps)


def patharrow(ps, *args):
	# the shaft will be itself an assembly of shafts
	# args is a quad path
	global sw
	global hw
	if len(args) == 1:
		A = args[0]
	else:
		A = args
	# first translate args into control points
	p0 = A[0][0]
	v0 = A[0][1]
	a = [p0]
	n = len(A)
	# args = array of [point, direction]
	for i in range(1, n):
		p2 = A[i][0]
		v2 = A[i][1]
		# control point p1 = = intersection of two lines
		p1 = Arrow.intersection(p0, v0, p2, v2)
		a.append(p1)
		a.append(p2)
		p0 = p2
		v0 = v2
	head = PlainHead()
	ell = head.length
	shaft = QuadShaft(ell, a)
	tail = StubTail()
	a = Arrow(head, shaft, tail)
	a.mkpath(ps)


