"""
	The basic idea in a BSP is to partition 
	the figure and space into convex sets, with each
	fragment completely contained in one of the convex sets.
	in addition is the idea of a tree-like partition.

"""

# --- the minimal object with location --------------

class Point:
	# c = centre, s = side
	def __init__(self, c):
		self.location = c

	def __str__(self):
		return str(self.location)

	def __getitem__(self, i):
		return self.location[i]

# P, Q are arrays of objects with locations in R^n
# inverse lexicographical order: last coordinates first
def comp(P, Q):
	p = P.location; q = Q.location
	# p, q now arrays of numbers
	n = len(p)
	for i in range(n):
		k = n-1-i
		if p[k] < q[k]: return -1
		elif p[k] > q[k]: return 1
	return 0

def split(P, n):
	x = P[0].location[n]
	ell = len(P)
	for i in range(1, ell):
		y = P[i].location[n]
		if x < y:
			return [ P[0:i], P[i:] ]
	return [ P, [] ]

# nodes of the tree at the end are [ P ] and [ f, back, front ]
# where S is an array of items with location
# f = hyperplane equation, back and front are nodes
# but during this program [ S, n ] also allowed, S = array of locations

def pointstree(P):
	# P as above
	P = sorted(P, comp)
	dim = len(P[0].location)
	root = [P, dim-1]
	stack = [ root ]
	while len(stack) > 0:
		node = stack.pop()
		# node = [ S, n ]
		# S = array of locations, n = index to sort by
		S = node[0] # a sorted array of locations
		ell = len(S) # we know ell > 0
		n = node[1] # all x_{i} for x in P and i < n are the same
		# print "popped:", toStr(S), ", ", n
		# if ell = 1 or n < 0, popping node makes it a true node [ S ]
		if ell == 1: # a single point 
			node.pop()
			node[0] = S[0]
		elif n < 0: # S is ell copies of the same point: reduce it to one copy?
			node.pop()
			""" for i in range(1,ell):
				S.pop() """
			node[0] = S[0]
			for i in range(1,ell):
				node.append(S[i])
		else: # try splitting according to character [n]
			s = split(S, n)
			if len(s[1]) == 0:
				node[1] = n-1 
				stack.append(node)
			else:
				# make a new node, plane in between s[0], s[1]
				S0 = s[0]; S1 = s[1]
				x0 = S0[0].location[n]; x1 = S1[0].location[n]
				# x0 = common value of [n] for x in s[0]
				# x1 = its value for first point in s[1]
				h = [0]*(dim+1)
				h[n] = 1; h[dim] = -(x0+x1)/2.0
				back = [ S0, n-1 ] # all the coordinates x_i of back are the same for i >= n
				front = [ S1, n ]
				stack.append(front)
				stack.append(back)
				node[0] = h
				node[1] = back
				node.append(front)
	return root

# -------------------------------------------------------------

def equationString(v):
	s = ""
	n = len(v)-1
	empty = True
	for i in range(n):
		if v[i] == 0:
			pass
		else:
			if not empty and v[i] > 0:
				s += "+"
			if v[i] == 1:
				s += "x_{" + str(i) + "}" 
			elif v[i] == -1:
				s += "-x_{" + str(i) + "}"
			else:
				s += str(v[i]) + "x_{" + str(i) + "}"
	s += " = " + str(-v[n])
	return s

def printtree(tree, k):
	if k > 0:
		print
	spc = ""
	for i in range(k):
		spc += "  "
	print spc,
	if len(tree) == 1: # [ P ]
		print spc, tree[0], "*",
	elif len(tree) == 2: # [ S, n ]
		print spc, tree[0], ":", tree[1], "*"
	else: # [ f, back, front ]
		print spc, equationString(tree[0]), ":",
		printtree(tree[1], k+1), ":",
		printtree(tree[2], k+1), "*"

# prints an arbitrary array, item by item
def toStr(P):
	s = "[ "
	for i in range(len(P)):
		s += str(P[i])
		s += ", "
	return s + "]"

# --------------------------------------------

"""

import random

random.seed(1)

N = 5
M = 16
dim = 3
p = []

for i in range(N):
	L = []
	for j in range(dim):
		L.append(random.randint(0, M))
	p.append(Point(L))

tree = pointstree(p)

printtree(tree, 0)

"""


