

from PSMatrix import PSMatrix

import math

class CoordinateSystem:

	def tm(self):
		return self.tm

	# vectors are columns
	def transform(self, *args):
		if len(args) == 2:
			x = args[0]; y = args[1]
		else: # len = 1
			x = args[0][0]; y = args[0][1]
		tm = self.tm
		X = tm[0]*x + tm[2]*y + tm[4]
		Y = tm[1]*x + tm[3]*y + tm[5]
		return [ X, Y ]

	# r = relative: acts on vectors
	def rtransform(self, V):
		dx = V[0]
		dy = V[1]
		tm = self.tm
		dX = tm[0]*dx + tm[2]*dy
		dY = tm[1]*dx + tm[3]*dy
		return [ dX, dY ]

	def inversetm(self):
		tm = self.tm
		d = float(tm[0]*tm[3]-tm[1]*tm[2])
		x = -tm[4]; y = -tm[5]
		B = [ tm[3]/d, -tm[1]/d, -tm[2]/d, tm[0]/d ]
		X = B[0]*x + B[2]*y; Y = B[1]*x + B[3]*y
		B.append(X); B.append(Y)
		return B

	# v = vector
	def itransform(self,  P):
		x = P[0]
		y = P[1]
		itm = self.inversetm()
		X = itm[0]*x + itm[2]*y + itm[4]
		Y = itm[1]*x + itm[3]*y + itm[5]
		return [ X, Y ]
		

	def __str__(self):
		tm = self.tm
		s = "["
		for i in range(6):
			s += " " + fstr(tm[i])
		s += " ] ("
		s += str(self.linewidth)
		s += ")"
		return(s)
		
	# b = array 
	def atransform(self, b):
		a = self.tm
		c = PSMatrix.concat(a, b)
		for i in range(6):
			self.tm[i] = c[i]
		
	# b = array 
	def ltransform(self, b):
		a = self.tm
		c = PSMatrix.lconcat(a, b)
		for i in range(4):
			self.tm[i] = c[i]
				
	def translate(self, *args):
		if len(args) == 1:
			v = args[0]
			x = v[0]
			y = v[1]
		else:
			x = args[0]
			y = args[1]
		tm = self.tm
		tm[4] += (x*tm[0] + y*tm[2])
		tm[5] += (x*tm[1] + y*tm[3])
		
	def scale(self, *args):
		if len(args) == 1:
			a = float(args[0])
			b = a
		else:
			a = float(args[0])
			b = float(args[1])
		tm = self.tm
		tm[0] *= a
		tm[1] *= a
		tm[2] *= b
		tm[3] *= b
	
	def rotate(self, A):
		c = math.cos(A)
		s = math.sin(A)
		tm = self.tm
		x = tm[0]; y = tm[2]
		t0 =  x*c + y*s
		t2 = -x*s + y*c
		tm[0] = t0
		tm[2] = t2
		x = tm[1]; y = tm[3]
		t1 =  x*c + y*s
		t3 = -x*s + y*c
		tm[1] = t1
		tm[3] = t3

