#!/usr/bin/python

from Fstr import fstr

from DviReader import *
import DeviceFont
import FindResource
from Tex import TexEnv
from PysCmdDevice import PysCmdDevice
import Tex
import Eigs2d 

from Canvas import *

""" from StringInsert import StringInsert
from Arc import *
from Matrix import Matrix
from Tex import * # imports Canvas, too
from StringInsert import *
from Graphics import *
from CMRFontDict import *
from StdFontDict import * """

import math
import sys

"""
	init arguments:

	Currently:
		w, h
		llx, lly, urx, ury - filename = present file + .eps
		extension, w, h
		extension, llx, lly, urx, ury - filename = present file + extension
		filename, w, h
		filename, llx, lly, urx, ury: 

	Future: 
		filename, no dims - bbox found by computation: init() possible

	if args[0] is a string: filename or extension, rest = 2 or 4 numbers
	else: args[0] = a number, 2 or 4 args
	
"""

class PageData:
	# level = level of corresponding graphics stack
	def __init__(self, cmd, level):
		c = []
		for i in range(len(cmd)):
			c.append(cmd[i])
		self.cmd = c
		self.level = level

""" Currently, PiScript performs two fairly separate tasks: 
(1) build a canvas
(2) produce output, through the device 
So the front end and back-end are a bit confounded. 

Well, maybe not.  All real graphics gets passed directly to Canvas. 
This has as extras only * bounding box, * page structure, * output.  
I.e. a Canvas ought to be no more than one graphics item,
no more than one page.  """

class PiScript(Canvas):
	# FIXME: extension determines kind of output (now .eps, .ps & later .pdf)
	# so: currently the device is a separate argument of init(), but *ought* to be 
	# determined by the arguments
	# currently the only device is PSExec

	# mode 0 = radians, 1 = degrees
	def __init__(self, device, *Args):
		args = []
		for a in Args:
			args.append(a)
		# copy Args to args so args can be read as a stack and shrunk
		cliptobbox = True
		if type(args[-1]) == str:
			if args[-1] == "noclip":
				cliptobbox = False
			args.pop()
		# --- sets filename, extension (.ps or .eps [default]) & bounding box ------------
		if len(args) > 0:
			a = args[0]
			if type(a) == str:
				if a == ".ps" or a == ".eps":
					self.ext = a
					filename = sys.modules['__main__'].__file__
					filename = filename[:-3] # drop the .py
				elif a[-3:] == ".ps":
					filename = a[:-3] # drop the .ps
					self.ext = ".ps"
				elif a[-4:] == ".eps":
					filename = a[:-4] # drop the .eps
					self.ext = ".eps"
				else:
					filename = a 
					self.ext = ".eps" # default extension
				if len(args) == 3:
					self.llx = 0
					self.lly = 0
					self.urx = int(args[1])
					self.ury = int(args[2])
				else: # len = 5
					self.llx = int(args[1])
					self.lly = int(args[2])
					self.urx = int(args[3])
					self.ury = int(args[4])
			else: # a number
				filename = sys.modules['__main__'].__file__
				if filename[-3:] == ".py":
					filename = filename[:-3] # drop the .py
				else:
					print "Expecting source file with .py extension"
					print "Exiting ... "
					sys.exit(1)
				self.ext = ".eps"
				if len(args) == 2:
					self.llx = 0
					self.lly = 0
					self.urx = int(args[0])
					self.ury = int(args[1])
				else: # len = 4
					self.llx = int(args[0])
					self.lly = int(args[1])
					self.urx = int(args[2])
					self.ury = int(args[3])
		else:
			print "No arguments for init!"
			print "Exiting ... "
			sys.exit(1)
		# -------------------------------------------------------------------

		Canvas.__init__(self)
		self.bbox = [self.llx, self.lly, self.urx, self.ury]

		# -------------------------------------------------------

		print "Initializing " + filename + self.ext
		self.filename = filename

		self.device = device;

		# -------------------------------------------------------
		# print "Bounding box: ", [ self.llx, self.lly, self.urx, self.ury ]
		print
		w = self.urx-self.llx
		h = self.ury-self.lly
		self.width = w
		self.height = h
		device.setbbox([ self.llx, self.lly, self.urx, self.ury ])
		# ------------------------------------------------------
		self.settexenv("TexConfig")
		self.insertno = 0 # for labeling saved inserts
		# ------------------------------------------------------

		self.pageno = 0
		self.pagestring = ""
		self.pagestack = [ PageData( [], self.glevel) ]
		self.beginpage()

		""" self.setlinewidth(1)
		self.setlinecap(0)
		self.setlinejoin(0)
		self.setdash([], 0)
		self.setmiterlimit(10.0)
		self.setcolor([0,0,0]) """

		if cliptobbox:
			self.newpath()
			self.boundedbox(self.bbox)
			self.clip()

		self.lablist = []
		self.endblock = True

	# =================================================================

	def currentcenter(self):
		gs = self.cgs()
		b = self.bbox
		x = 0.5*(b[0]+b[2])
		y = 0.5*(b[1]+b[3])
		return gs.transform((x, y))

	def currentbbox(self):
		b = self.bbox
		gs = self.cgs()
		ll = gs.transform((b[0], b[1]))
		lr = gs.transform((b[2], b[1]))
		ur = gs.transform((b[2], b[3]))
		ul = gs.transform((b[0], b[3]))
		return [ ll, lr, ur, ul ]
	
	# ===================================================================================

	def center(self):
		b = self.bbox
		x = 0.5*(b[0]+b[2])
		y = 0.5*(b[1]+b[3])
		self.translate(x, y)

	def beginpage(self, *args):
		self.pagestring += "("
		opd = self.pagestack[-1]
		pd = PageData(opd.cmd, opd.level)
		self.pagestack.append(pd)
		self.cmd = pd.cmd
		self.gsave()

	def endpage(self):
		n = len(self.pagestack)
		if self.endblock and n == 2:
			print "Oops!  You are attempting to finish prematurely!"
			print "Now exiting ... !"
			sys.exit(1)
		# --- check graphics stack --------------------------------------
		ell1 = self.glevel
		ell0 = self.baselevel()+1
		ell = ell1-ell0
		if not ell == 0:
			print
			if ell < 0:
				print "WARNING! ====== There is an extra grestore! ====== "
			else: # ell > 0:
				print
				if ell > 1:
					print "WARNING! ====== There are " + str(ell) + " too many gsaves! ====== "
				else: # ell = 1
					print "WARNING! ====== There is an extra gsave! ====== "
			print "=============== Restoring to default coordinates! ====== "
		for i in range(ell+1):
			self.grestore()

		# --------------------------------------------------------------
		# something funny we did about not needing begin page?
		if self.pagestring[-1] == "(":
			self.device.writePage( self )
			self.pageno += 1
			print "[" + str(self.pageno) + "]",
		self.pagestring += ")"
		self.pagestack.pop()
		pd = self.pagestack[-1]
		self.cmd = pd.cmd

	def Inv(M):
		d = float(M[0]*M[3]-M[1]*M[2])
		return ( M[3]/d, -M[1]/d, -M[2]/d, M[0]/d )
	inv = staticmethod(Inv)

	def Transform(M, P):
		return (M[0]*P[0]+M[1]*P[1], M[2]*P[0]+M[3]*P[1])
	Transform = staticmethod(Transform)

	def finish(self):
		self.endblock = False
		self.endpage()
		# write the label file
		if len(self.lablist) > 0:
			# All this could be done when the thing is placed!  
			# Only depends on data available then.
			# so: pass just the 
			labelfile = open(self.filename + ".lab", "w")
			for ell in self.lablist:
				string = ell[0]
				matrix = ell[1]
				location = ell[2]
				origin = ell[3]
				# print "original origin = ", origin
				# print "original location =", location
				# print "matrix =", matrix
				M = [matrix[0], matrix[2], matrix[1], matrix[3]] # rotation matrix if just rotated
				[k1, a, k2] = Eigs2d.kak(M)
				A1 = math.atan2(k1[2], k1[0])
				A1 = (180.0/math.pi)*A1
				R1 = "\\RotateBox{" + str(A1) + "}"

				A2 = math.atan2(k2[2], k2[0])
				A2 = (180.0/math.pi)*A2
				R2 = "\\RotateBox{" + str(A2) + "}"

				S = "\\scalebox{" + str(a[0]) + "}[" + str(a[3]) + "]"

				matrixstr = R1 + "{" + S + "{" + R2 + "{" + str(ell[0]) + "}}}"

				(o0, o1) = PiScript.Transform(M, (origin[0], origin[1]))
				x = matrix[4] + o0; y = matrix[5] + o1
				location[0] += x 
				location[1] += y 
				
				# location is in Adobe points
				label = matrixstr + " [Bl] at " + str((72.27/72)*location[0]) + " " + str((72.27/72)*location[1]) + "\n"
				labelfile.write("\\pinlabel* " + label)
			labelfile.close()
		else: # delete the lab file if it exists
			os.system("rm -f " + self.filename + ".lab")
		finalout = open( self.filename+self.ext, "w" )
		self.device.finish( finalout, toEPS = (self.ext=='.eps') )
		finalout.close()

	""" this is a duplicate!
	# FIXME: WAC, return in currentcoords
	def currentbbox(self):
		gs = self.cgs()
		return ( 
			gs.itransform((self.llx, self.lly)), 
			gs.itransform((self.urx, self.lly)), 
			gs.itransform((self.urx, self.ury)), 
			gs.itransform((self.llx, self.ury))
		)
	"""

	def baselevel(self):
		opd = self.pagestack[-2]
		return opd.level

	def gsave(self):
		pd = self.pagestack[-1]
		pd.level += 1
		Canvas.gsave(self)

	def grestore(self):
		ell0 = self.baselevel()
		ell1 = self.glevel
		if ell1 > ell0:
			Canvas.grestore(self)
			pd = self.pagestack[-1]
			pd.level -= 1
		else:
			print "WARNING! too many grestores!"
			

	# ============================================================

	def settexprefix(self, s):
		self.texenv.setprefix(s)

	def settexmacros(self, s):
		self.texenv.setmacros(s)

	def settexpostfix(self, s):
		self.texenv.setpostfix(s)

	def settexcommand(self, s):
		self.texenv.setcommand(s)

	def settexsave(self, save):
		self.texenv.setsave(save)

	"""
		Either one argument or 3 +                                                                                          
		if 3, then strings = prefix, command, postfix
		if 1, then
			"plain"
			"latex"
			None
			filename like "PSTexConfig": checks locally, then goes global
	"""

	def settexenv(self, *args):
		print "TEX = ", args[0]
		if len(args) == 1: # arg[0] = "plain" or "latex" or TexConfig file name
			if args[0] == "plain":
				cfgfile = "PlainTexConfig"
			elif args[0] == "latex":
				cfgfile = "LaTexConfig"
			else:
				cfgfile = args[0]
			# so now cgffile is the prefix of a TexConfig file
			# -------------------------------------------------------
			# check current directory, 
			# then personal cfg dir, 
			# then normal PYS directory
			cfg = os.getcwd()
			cfgpathname = os.path.join(cfg, cfgfile + ".py")
			if os.path.exists(cfgpathname):
				print "importing local " + cfgpathname
				
				f = open(cfgpathname, "r")
				r = f.read()
				exec(r)
				self.texenv = getTexEnv()

				"""
				print "------------------------"
				exec "import " + cfgfile
				self.texenv = eval(cfgfile + ".getTexEnv()")
				print "------------------------" 
				"""

				return
			# -------------------------------------------------------
			if os.environ.has_key("LOCALPISCRIPTDIR"):
				cfg = os.environ["LOCALPISCRIPTDIR"]
				home = os.getenv("HOME")
				import re
				m = re.search("(\$HOME)", cfg)
				X = re.subn("\$HOME", home, cfg)
				cfg = X[0]
				if cfg:
					cfgpathname = os.path.join(cfg, "configs", cfgfile + ".py")
					if os.path.exists(cfgpathname):
						print "importing home " + cfgpathname

						f = open(cfgpathname, "r")
						r = f.read()
						exec(r) 
						self.texenv = getTexEnv()

						"""
						exec "import configs." + cfgfile 
						print "importing " + cfgpathname
						self.texenv = eval("configs." + cfgfile + ".getTexEnv()")
						"""
						return
			# -------------------------------------------------------
			cfg = os.path.abspath(os.path.dirname(__file__))
			cfgpathname = os.path.join(cfg, cfgfile + ".py")
			if os.path.exists(cfgpathname):
				print "importing global " + cfgpathname

				f = open(cfgpathname, "r")
				r = f.read()
				exec(r) 
				self.texenv = getTexEnv()
				"""
				exec "import piscript." + cfgfile 
				print "importing " + cfgpathname
				self.texenv = eval("piscript." + cfgfile + ".getTexEnv()")
				"""
				return
			# -------------------------------------------------------
			print "WARNING! TEX configuration ", cfgfile, " not found!"
			print "?", cfgpathname
			print "Aborting mission!"
			import sys
			sys.exit(1)
		else: # prefix, command, postfix
			self.texenv = TexEnv(*args)

	def stringinsert(self, s):
		c = StringInsert(s)
		return c

	# orig args = (texstring), (texstring, save/not, pin=T/F)
	def texinsert(self, texstring, save=None, pin=False):
		s = save
		p = pin
		""" texstring = args[0]
		if len(args) == 1:
			if self.texenv.save == True:
				s = "tmp-" + self.filename + "-" + str(self.insertno)  # tex label
				self.insertno += 1
		else:
			if type(args[1]) == bool and args[1] == True:
				s = "tmp-" + self.filename + "-" + str(self.insertno)  # tex label
				self.insertno += 1
			else: # string = name to be saved by
				s = "tmp-" + self.filename + "-" + args[1] # tex label """

		# The 792 is because TEX is set up for North America ... ?  
		# If in Europe?  WAC: FIXME?
		tr = Tex.TexRunner( self.texenv,  texstring, PysCmdDevice(792, self.currentcolor()), save=s)
		T = tr.device.pages[0] # a TexInsert = special Canvas
		T.pinned = pin
		T.texstring = texstring
		return T

	def place(self, canvas, *args):
		s = Canvas.place(self, canvas, *args)
		if s: # place returns the tex label if pinned, else None
			if len(args) == 0:
				P = (0, 0)
			if len(args) == 1:
				P = args[0]
			elif len(args) == 2:
				P = (args[0], args[1])
			gs = self.cgs() 
			v = gs.transform(P)
			# print "label =", s
			# print "matrix =", canvas.m
			m = [ x for x in canvas.m ] # save a copy, original might change later
			origin = [ x for x in canvas.origorigin ]
			# print "v =", v
			self.lablist.append([s, m, v, origin])
		else:
			pass



